/* NVTV actions -- Dirk Thierbach <dthierbach@gmx.de>
 *
 * This is open software protected by the GPL. See GPL.txt for details.
 *
 * Header: Intermediate layer. Contains tv actions, X actions and card
 *   access.
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/extensions/xf86vmode.h>

#include "backend.h"
#include "debug.h"
#include "xfree.h"
#include "nv_type.h"
#include "tv_nv.h"

#include "actions.h"

/* -------- Print actions -------- */

void print_nv_regs (TVNvRegs *m)
{
  printf ("*** CRT Registers\n");
  printf ("  HDisplay    : %i,\n", m->HDisplay);
  printf ("  HSyncStart  : %i,\n", m->HSyncStart);   
  printf ("  HSyncEnd    : %i,\n", m->HSyncEnd);     
  printf ("  HTotal      : %i,\n", m->HTotal);       
  printf ("  VDisplay    : %i,\n", m->VDisplay);     
  printf ("  VSyncStart  : %i,\n", m->VSyncStart);   
  printf ("  VSyncEnd    : %i,\n", m->VSyncEnd);     
  printf ("  VTotal      : %i,\n", m->VTotal);       
}

void print_crt_regs (TVCrtcRegs *m, CardType card)
{
  switch (card)
  {
  case CARD_NVIDIA:
    print_nv_regs (&m->nv);
    break;
  default:
    break;
  }
}

void print_bt_regs (TVBtRegs *b)
{
  printf ("*** Brooktree Registers\n");
  printf ("  hsynoffset   : %i,\n", b->hsynoffset);
  printf ("  vsynoffset   : %i,\n", b->vsynoffset);
  printf ("  hsynwidth    : %i,\n", b->hsynwidth);
  printf ("  vsynwidth    : %i,\n", b->vsynwidth);
  printf ("  h_clko       : %i,\n", b->h_clko);
  printf ("  h_active     : %i,\n", b->h_active);
  printf ("  hsync_width  : %i,\n", b->hsync_width);
  printf ("  hburst_begin : %i,\n", b->hburst_begin);
  printf ("  hburst_end   : %i,\n", b->hburst_end);
  printf ("  h_blanko     : %i,\n", b->h_blanko); 
  printf ("  v_blanko     : %i,\n", b->v_blanko); 
  printf ("  v_activeo    : %i,\n", b->v_activeo);
  printf ("  h_fract      : %i,\n", b->h_fract);  
  printf ("  h_clki       : %i,\n", b->h_clki);   
  printf ("  h_blanki     : %i,\n", b->h_blanki); 
  printf ("  v_linesi     : %i,\n", b->v_linesi); 
  printf ("  v_blanki     : %i,\n", b->v_blanki); 
  printf ("  v_activei    : %i,\n", b->v_activei);
  printf ("  v_scale      : %i,\n", b->v_scale);  
  printf ("  pll_fract    : %i,\n", b->pll_fract);
  printf ("  pll_int      : %i,\n", b->pll_int);  
  printf ("  sync_amp     : %i,\n", b->sync_amp); 
  printf ("  bst_amp      : %i,\n", b->bst_amp);  
  printf ("  mcr          : %i,\n", b->mcr);      
  printf ("  mcb          : %i,\n", b->mcb);      
  printf ("  my           : %i,\n", b->my);       
  printf ("  msc          : %li,\n", b->msc); 
}

void print_ch_regs (TVChRegs *r)
{
  printf ("*** Chrontel Registers\n");
  printf ("  dmr_ir    : %i,\n", r->dmr_ir);
  printf ("  dmr_vs    : %i,\n", r->dmr_vs);
  printf ("  dmr_sr    : %i,\n", r->dmr_sr);
  printf ("  sav       : %i,\n", r->sav);
  printf ("  hpr       : %i,\n", r->hpr);
  printf ("  vpr       : %i,\n", r->vpr);
  printf ("  pll_m     : %i,\n", r->pll_m);
  printf ("  pll_n     : %i,\n", r->pll_n);
  printf ("  pllcap    : %i,\n", r->pllcap);
  printf ("  dacg      : %i,\n", r->dacg);
  printf ("  fsci      : %li,\n", r->fsci);
  printf ("  ffr_fc    : %i,\n", r->ffr_fc);
  printf ("  ffr_fy    : %i,\n", r->ffr_fy);
  printf ("  ffr_ft    : %i,\n", r->ffr_ft);
  printf ("  blr       : %i,\n", r->blr);
  printf ("  ce        : %i,\n", r->ce);
  printf ("  vbw_flff  : %i,\n", r->vbw_flff);  
  printf ("  vbw_cvbw  : %i,\n", r->vbw_cvbw);  
  printf ("  vbw_cbw   : %i,\n", r->vbw_cbw);   
  printf ("  vbw_ypeak : %i,\n", r->vbw_ypeak); 
  printf ("  vbw_ysv   : %i,\n", r->vbw_ysv);   
  printf ("  vbw_ycv   : %i,\n", r->vbw_ycv);    
}

void print_ph_regs (TVPhRegs *r)
{
  printf ("adwhs  : %i,\n", r->adwhs);    
  printf ("adwhe  : %i,\n", r->adwhe);    
  printf ("xofs   : %i,\n", r->xofs);     
  printf ("xpix   : %i,\n", r->xpix);     
  printf ("xinc   : %i,\n", r->xinc);     
  printf ("hlen   : %i,\n", r->hlen);     
  printf ("fal    : %i,\n", r->fal);      
  printf ("lal    : %i,\n", r->lal);      
  printf ("yinc   : %i,\n", r->yinc);     
  printf ("yskip  : %i,\n", r->yskip);    
  printf ("yofso  : %i,\n", r->yofso);    
  printf ("yofse  : %i,\n", r->yofse);    
  printf ("ypix   : %i,\n", r->ypix);     
  printf ("yiwgto : %i,\n", r->yiwgto);   
  printf ("yiwgte : %i,\n", r->yiwgte);   
  printf ("pcl    : %li,\n", r->pcl);    
  printf ("fsc    : %li,\n", r->fsc);    
  printf ("idel   : %i,\n", r->idel);     
  printf ("bs     : %i,\n", r->bs);       
  printf ("be     : %i,\n", r->be);       
  printf ("bsta   : %i,\n", r->bsta);     
  printf ("blckl  : %i,\n", r->blckl);    
  printf ("blnnl  : %i,\n", r->blnnl);    
  printf ("chps   : %i,\n", r->chps);     
  printf ("gy     : %i,\n", r->gy);       
  printf ("gcd    : %i,\n", r->gcd);      
  printf ("bcy    : %i,\n", r->bcy);      
  printf ("bcu    : %i,\n", r->bcu);      
  printf ("bcv    : %i,\n", r->bcv);      
  printf ("ccrs   : %i,\n", r->ccrs);     
  printf ("gainu  : %i,\n", r->gainu);    
  printf ("gainv  : %i,\n", r->gainv);    
}

void print_tv_regs (TVEncoderRegs *c, TVChip type)
{
  switch (type & TV_ENCODER) 
  {
    case TV_BROOKTREE:
    case TV_CONEXANT:
      print_bt_regs (&c->bt); 
      break;
    case TV_CHRONTEL:
      print_ch_regs (&c->ch); 
      break;
    case TV_PHILIPS:
      print_ph_regs (&c->ph); 
      break;
    default:
      break;
  }
}

/* -------- X Actions -------- */

/* Minimum vidmode extension version required */
#define MINMAJOR 0
#define MINMINOR 5

inline int sqr (int arg)
{
  return arg * arg;
}

Bool has_vidmode (Display *display)
{
  static int vidmode = 0;

  int MajorVersion, MinorVersion;
  int EventBase, ErrorBase;

  while (vidmode == 0) {
    if (!XF86VidModeQueryVersion (display, &MajorVersion, &MinorVersion))
    {
      fprintf(stderr, "Unable to query video extension version\n");
      vidmode = -1;
      break;
    }
    if (!XF86VidModeQueryExtension(display, &EventBase, &ErrorBase)) {
      fprintf(stderr, "Unable to query video extension information\n");
      vidmode = -1;
      break;
    }
    /* Fail if the extension version in the server is too old */
    if (MajorVersion < MINMAJOR || 
	(MajorVersion == MINMAJOR && MinorVersion < MINMINOR)) {
      fprintf(stderr,
	      "Xserver is running an old XFree86-VidModeExtension version"
	      " (%d.%d)\n", MajorVersion, MinorVersion);
      fprintf(stderr, "Minimum required version is %d.%d\n",
	      MINMAJOR, MINMINOR);
      vidmode = -1;
      break;
    }
    vidmode = 1;
    break;
  }
  if (vidmode > 0) return TRUE; else return FALSE;
}

/* get all modelines only once, so we don't have to free them */

int vidmodelines (Display *display, int screen, XF86VidModeModeInfo ***info)
{
  static XF86VidModeModeInfo **modesinfo = NULL; 
  static int modecount = 0;

  if (!modesinfo) {
    XF86VidModeGetAllModeLines (display, screen, &modecount, &modesinfo);
  }
  *info = modesinfo;
  return modecount;
}

void switch_vidmode (Display *display, int screen, int res_x, int res_y)
{
  XF86VidModeModeInfo **modesinfo;
  int modecount;
  int i;
  int dist;
  int best_dist = -1;
  int best_index = -1;

  DPRINTF ("vidmode %i %i\n", res_x, res_y);
  if (!has_vidmode (display)) return;
  modecount = vidmodelines (display, screen, &modesinfo);
  for (i = 0; i < modecount; i++) {
    dist = sqr(modesinfo [i]->hdisplay - res_x) + 
           sqr(modesinfo [i]->vdisplay - res_y);
    if (modesinfo [i]->hdisplay < res_x ||
        modesinfo [i]->vdisplay < res_y) {
      dist = -1;
    }
    if (dist >= 0 && (dist < best_dist || best_dist == -1)) {
      best_dist = dist;
      best_index = i;
    }
  }
  if (best_dist >= 0) {
    DPRINTF ("vidmode switch %i\n", best_index);
    XF86VidModeSwitchToMode (display, screen, modesinfo [best_index]);
  }
}

/*
 * Find X video mode with given resolution
 */

Bool find_vidmode (Display *display, int screen, int res_x, int res_y, 
  TVCrtcRegs *crt, make_vidmode make)
{
  XF86VidModeModeInfo **modesinfo;
  int modecount;
  int i;

  if (!has_vidmode (display)) return FALSE;
  modecount = vidmodelines (display, screen, &modesinfo);
  for (i = 0; i < modecount; i++) {
    if (modesinfo [i]->hdisplay == res_x && modesinfo [i]->vdisplay == res_y) 
      break;
  }
  if (i >= modecount) return FALSE;
  make (modesinfo [i]->hdisplay, modesinfo [i]->hsyncstart, 
	modesinfo [i]->hsyncend, modesinfo [i]->htotal,    
	modesinfo [i]->vdisplay, modesinfo [i]->vsyncstart,
	modesinfo [i]->vsyncend, modesinfo [i]->vtotal, 
	modesinfo [i]->dotclock, crt);
  return TRUE;
}

/*
 * get current X video mode
 */

Bool get_vidmode (Display *display, int screen, int *res_x, int *res_y,
  TVCrtcRegs *crt, make_vidmode make)
{
  XF86VidModeModeLine modeline;
  int dotclock;

  if (!has_vidmode (display)) return FALSE;
  XF86VidModeGetModeLine (display, screen, &dotclock, &modeline);
  if (res_x) *res_x = modeline.hdisplay;
  if (res_y) *res_y = modeline.vdisplay;
  make (modeline.hdisplay, modeline.hsyncstart, 
	modeline.hsyncend, modeline.htotal,    
	modeline.vdisplay, modeline.vsyncstart,
	modeline.vsyncend, modeline.vtotal, 
	dotclock, crt);
  return TRUE;
}

void center_window (Display *display, int screen, Window window,
  int res_x, int res_y)
{
  Window root;
  int x, y;
  unsigned width, height, borderw, depth;
  unsigned disp_w, disp_h;
  int screen_w, screen_h;
  int rx, ry;
  Window junkwin;
  int junkint;
  XF86VidModeModeLine modeline;

  DPRINTF ("center 0x%lx %i,%i\n", window, res_x, res_y);
  if (window == None || !has_vidmode (display)) return;
  XGetGeometry (display, window, &root, &x, &y, &width, &height, 
		&borderw, &depth);
  (void) XTranslateCoordinates (display, window, root, 
				0, 0, &rx, &ry, &junkwin);
  if (res_x >= 0 && res_y >= 0) {
    disp_w = res_x;
    disp_h = res_y;
  } else {
    XF86VidModeGetModeLine (display, screen, &junkint, &modeline);
    if (modeline.privsize > 0) XFree(modeline.private);
    disp_w = modeline.hdisplay;
    disp_h = modeline.vdisplay;
  }
  screen_w = DisplayWidth (display, screen);
  screen_h = DisplayHeight (display, screen);
  x = rx + ((int) width - (int) disp_w) / 2;
  y = ry + ((int) height - (int) disp_h) / 2;
  if (x < 0) x = 0;
  if (y < 0) y = 0;
  if (x > screen_w) x = screen_w;
  if (y > screen_h) y = screen_h;
  XF86VidModeSetViewPort (display, screen, x, y);
  XWarpPointer (display, None, window, 0, 0, 0, 0, 
		width / 2, height / 2);
  /* FIXME: Better, disable the pointer by setting an empty pixmap
     or grabbing it, and restore image on leave event or on the next
     mouse event. Will work only in interactive (gui) mode! */
}

/* -------- dsimple.c -------- */

/* This is are (slightly modified) routines from programs/xlsfonts/dsimple.c 
 * in the XFree sources, written by Mark Lillibridge.
 */

#define Fatal_Error(X) { fprintf(stderr, X); exit (1); }

Window Select_Window (Display *dpy, int screen)
{
  int status;
  Cursor cursor;
  XEvent event;
  Window target_win = None, root = RootWindow(dpy,screen);
  int buttons = 0;

  /* Make the target cursor */
  cursor = XCreateFontCursor(dpy, XC_crosshair);

  /* Grab the pointer using target cursor, letting it room all over */
  status = XGrabPointer(dpy, root, False,
			ButtonPressMask|ButtonReleaseMask, GrabModeSync,
			GrabModeAsync, root, cursor, CurrentTime);
  if (status != GrabSuccess) Fatal_Error("Can't grab the mouse.");

  /* Let the user select a window... */
  while ((target_win == None) || (buttons != 0)) {
    /* allow one more event */
    XAllowEvents(dpy, SyncPointer, CurrentTime);
    XWindowEvent(dpy, root, ButtonPressMask|ButtonReleaseMask, &event);
    switch (event.type) {
    case ButtonPress:
      if (target_win == None) {
	target_win = event.xbutton.subwindow; /* window selected */
	if (target_win == None) target_win = root;
      }
      buttons++;
      break;
    case ButtonRelease:
      if (buttons > 0) /* there may have been some down before we started */
	buttons--;
       break;
    }
  } 
  XUngrabPointer(dpy, CurrentTime);      /* Done with pointer */
  return(target_win);
}

/*
 * Window_With_Name: routine to locate a window with a given name on a display.
 *                   If no window with the given name is found, 0 is returned.
 *                   If more than one window has the given name, the first
 *                   one found will be returned.  Only top and its subwindows
 *                   are looked at.  Normally, top should be the RootWindow.
 */

Window Window_With_Name (Display *dpy, Window top, char *name)
{
  Window *children, dummy;
  unsigned int nchildren;
  int i;
  Window w = None;
  char *window_name;

  if (XFetchName(dpy, top, &window_name) && !strcmp(window_name, name))
    return(top);

  if (!XQueryTree(dpy, top, &dummy, &dummy, &children, &nchildren))
    return None;

  for (i=0; i<nchildren; i++) {
    w = Window_With_Name(dpy, children[i], name);
    if (w != None)
      break;
  }
  if (children) XFree ((char *)children);
  return w;
}
