/* NVTV main -- Dirk Thierbach <dthierbach@gmx.de>
 *
 * This file is part of nvtv, a tool for tv-output on NVidia cards.
 * 
 * nvtv is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * nvtv is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: main.c,v 1.1.1.1 2003/01/02 04:49:48 rshoff Exp $
 *
 * Contents:
 *
 * Main routine. Parse options, execute actions or call gui.
 *
 */

#include <stdio.h>
#include <getopt.h>
#include <string.h>
#include <unistd.h>
#include "xfree.h"

#include <X11/Xlib.h>
#ifdef HAVE_GTK
#include <gdk/gdkprivate.h>
#include "gui.h"
#endif

#include "backend.h"
#include "back_direct.h"
#include "back_nvidia.h"
#include "back_null.h"
#include "back_client.h"
#include "main.h"
#include "actions.h"

/* Debug: */

#include "tv_nv.h"

/* -------- Global options -------- */

Bool opt_no_root = FALSE;

Bool opt_nvdev = FALSE;

int opt_res_x = -1, opt_res_y = -1;
float opt_hoc = -1.0, opt_voc = -1.0;

int opt_head = -1;

char *opt_size = NULL;
char *opt_dpy = NULL;
char *opt_window_name = NULL;

BackAccessPtr back_access = NULL;  /* global backend */
BackCardPtr back_card = NULL;

TVChip opt_tv_chip = TV_NO_CHIP;
TVSystem opt_system = TV_SYSTEM_NONE;
TVConnect opt_connect = CONNECT_AUTO;
Window opt_window = None;

int opt_service_flags = 0;
int opt_service_mask = 0;

int opt_mode_flags = 0;
int opt_mode_mask = 0;

TVSettings opt_set;

static const char *short_options = "?12bcdf:ghmno:pr:s:tw:C:NPS:T:W:X";

static struct option long_options[] =
  {{"tv-bars",        no_argument,       NULL, 'b'},
   {"center",         no_argument,       NULL, 'c'},
   {"debug",          no_argument,       NULL, 'd'},
   {"set-flicker",    required_argument, NULL, 'f'},
   {"gui",            no_argument,       NULL, 'g'},
   {"help",           no_argument,       NULL, 'h'},
   {"tv-off",         no_argument,       NULL, 'm'},
   {"no-root",        no_argument,       NULL, 'n'},
   {"overscan",       required_argument, NULL, 'o'},
   {"print",          no_argument,       NULL, 'p'},
   {"resolution",     required_argument, NULL, 'r'},
   {"size",           required_argument, NULL, 's'},
   {"tv-on",          no_argument,       NULL, 't'},
   {"win-name",       required_argument, NULL, 'w'},
   {"display",        required_argument, NULL, 'D'},
   {"connector",      required_argument, NULL, 'C'},
   {"nvdev",          no_argument,       NULL, 'N'},
   {"probe",          no_argument,       NULL, 'P'},
   {"system",         required_argument, NULL, 'S'},
   {"chip",           required_argument, NULL, 'T'},
   {"win-id",         required_argument, NULL, 'W'},
   {"switch-mode",    no_argument,       NULL, 'X'},
   {"set",            required_argument, NULL, 300},
   {"sync",           no_argument,       NULL, '.'},
   {"no-xshm",        no_argument,       NULL, '.'},
   {"name",           required_argument, NULL, '.'},
   {"class",          required_argument, NULL, '.'},
   {NULL,             0,                 NULL, 0}
};

OptIntVal *opt_set_act = NULL;

#define SET_FIELD(f) addr:&opt_set.f

static OptIntDecl opt_set_list [] = {
  {"Contrast",        min:-100, max:100, SET_FIELD(contrast)},
  {"Saturation",      min:-100, max:100, SET_FIELD(saturation)},
  {"Brightness#",     min: -50, max: 50, SET_FIELD(brightness_sig)}, 
  {"Contrast#",       min: -50, max: 50, SET_FIELD(contrast_sig)},
  {"Saturation#",     min: -50, max: 50, SET_FIELD(saturation_sig)},
  {"Phase",           min: -60, max: 60, SET_FIELD(phase)},
  {"Hue",             min: -60, max: 60, SET_FIELD(hue)},
  {"Flicker",         min:   0, max:100, SET_FIELD(flicker)},
  {"AdaptFlicker",    min:   0, max:100, SET_FIELD(flicker_adapt)},
  {"LumaBandwidth",   min:   0, max:100, SET_FIELD(luma_bandwidth)},
  {"ChromaBandwidth", min:   0, max:100, SET_FIELD(chroma_bandwidth)},
  {"Sharpness",       min:   0, max:100, SET_FIELD(sharpness)},
  {"CrossColor",      min:   0, max:100, SET_FIELD(cross_color)},
  {NULL}
};

#define SET_FLAG(x) addr_val:&opt_##x##_flags, addr_mask:&opt_##x##_mask

static OptFlagDecl opt_flag_list [] = {
  {"Dualview",        mask: TV_DESC_DUALVIEW,      SET_FLAG(mode)},
  {"Monochrome",      mask: TV_DESC_MONOCHROME,    SET_FLAG(mode)},
  {"Non-Interlaced",  mask: TV_DESC_NONINTERLACED, SET_FLAG(mode)},
  {"Macrovision",     mask: TV_DESC_MACROVISION,   SET_FLAG(mode)},
  {"FreeCarrier",     mask: TV_DESC_CARRIER_LOCK,  SET_FLAG(mode)},
  {"Colorfix",        mask: TV_DESC_COLORFIX,      SET_FLAG(mode)},
  {"AdjustCursor",    mask: BACK_SERVICE_CURSOR,      SET_FLAG(service)},
  {"ViewportCursor",  mask: BACK_SERVICE_VIEW_CURSOR, SET_FLAG(service)},
  {"ViewportMonitor", mask: BACK_SERVICE_VIEW_MAIN,   SET_FLAG(service)},
  {NULL}
};

/* -------- Actions -------- */

typedef enum {
  ACTION_NONE    = 0,
  ACTION_PRINT   = 1,
  ACTION_GUI     = 2,
  ACTION_TVON    = 3,
  ACTION_TVOFF   = 4,
  ACTION_TVBARS  = 5,
  ACTION_DEBUG   = 6,
  ACTION_PROBE   = 7,
} ActionMain;

typedef enum {
  ACTION_CENTER  = (1 << 0),
  ACTION_XMODE   = (1 << 1),
} ActionAux;

/* -------- */

void usage (void)
{
  OptIntDecl *p;
  OptFlagDecl *r;
  int i;

  fprintf (stderr,
      "usage:  nvtv [-options ...]\n\n");
  fprintf (stderr,
      "where options include:\n");
  fprintf (stderr,
      "  -h --help             print this message\n");
  fprintf (stderr,	       
      "     --display hst:scrn X server to contact\n");
  fprintf (stderr,	       
      "  -p --print            calculate and print register values\n");
  fprintf (stderr,	       
      "  -m --tv-off           TV off, switch to monitor\n");
  fprintf (stderr,	       
      "  -t --tv-on            TV on\n");
  fprintf (stderr,	       
      "  -b --tv-bars          Color bars\n");
  fprintf (stderr,	       
      "  -r --resolution x,y   Resolution\n");
  fprintf (stderr,	       
      "  -o --overscan x,y     Overscan percentage (float)\n");
  fprintf (stderr,	       
      "  -s --size <size>      Size (for predefined modes)\n");
  fprintf (stderr,	       
      "  -S --system           NTSC,NTSC-J,PAL,PAL-60,PAL-NC,PAL-M,PAL-M60\n");
  fprintf (stderr,	       
      "  -C --connector        AUTO,FBAS,SVHS,BOTH,CONVERT\n");
  fprintf (stderr,
      "  -T --chip             BROOKTREE,CHRONTEL,CONEXANT,PHILIPS\n");
  fprintf (stderr,
      "  -W --win-id id        select window by numerical id\n");
  fprintf (stderr,
      "  -w --win-name name    select window by name\n");
  fprintf (stderr,
      "  -c --center           center selected window\n");
  fprintf (stderr,
      "  -X --switch-mode      switch X mode\n");
  fprintf (stderr,
      "  -1                    use first head\n");
  fprintf (stderr,
      "  -2                    use second head (if available)\n");
  fprintf (stderr,
      "  -P --probe            probe and print system information\n");
  fprintf (stderr,
      "  -N --nvdev            enable usage of /dev/nv* devices\n");
  fprintf (stderr,
      "  -g --gui              always use gui\n");
  fprintf (stderr,
      "  -n --no-root          (only for debugging)\n");
  fprintf (stderr,
      "  -d --debug            (only for debugging)\n");
  fprintf (stderr,             
      "     --set name:val     Specify setting\n");
  fprintf (stderr, "\nThe following settings are available:\n");
  for (i = 0, p = opt_set_list; p->name; i++, p++) 
  {
    fprintf (stderr, "  %-20s (%4i to %4i)", p->name, p->min, p->max);
    if (i % 2) {
      fprintf (stderr, "\n");
    } else {
      fprintf (stderr, "  ");
    }
  }
  if (i % 2) fprintf (stderr, "\n");
  for (i = 0, r = opt_flag_list; r->name; i++, r++) 
  {
    fprintf (stderr, "  %-20s (flag) %4s", r->name, "");
    if (i % 2) {
      fprintf (stderr, "\n");
    } else {
      fprintf (stderr, "  ");
    }
  }
  if (i % 2) fprintf (stderr, "\n");
  exit (1);
}

void printRegs (TVMode *mode, CardType card)
{
  printf ("*** Resolution %03i x %03i  Overscan %06.3f x %06.3f\n",
	  mode->spec.res_x, mode->spec.res_y, mode->spec.hoc, mode->spec.voc);
  print_tv_regs (&mode->regs.enc, opt_tv_chip);
  print_crt_regs (&mode->regs.crtc, card);
}

#include "tv_common.h"
#include "tv_i2c.h"

void testWr (I2CDevPtr dev, int addr)
{
  tvBusOk = TRUE;
  TVWriteBus (dev, addr, 0x00);
  printf ("%02X:%c ", addr, tvBusOk ? '+' : '-');
}

#include "tv_nv.h"

void doDebug (void)
{
  CardPtr card;
  I2CChainPtr chain;
  I2CDevPtr dev;
  I2CByte res;
  TVRegs r;
  int max;
  int i, j;

  if (! back_root_avail ()) {
    printf ("No root backend\n");
    exit (1);
  }
  card = back_root_init ();
  printf ("Root backend is open, card=%s\n", card->name);
  back_access->openCard (card);
  printf ("Card is open\n");
  if(!card->chips) {
    printf("No chips\n");
    exit(0);
  }
  chain = (I2CChainPtr) card->chips->private;
  dev = chain->dev;
  printf ("Detected chip addr=%02X type=%s (possibly wrong)\n",
          dev->SlaveAddr, card->chips->name);
#if 0
  switch (dev->SlaveAddr)
  {
    case 0x88:
      printf ("Assuming Philips\n");
      printf ("Testing writes:\n");
      testWr (dev, 0x00); testWr (dev, 0x1C); testWr (dev, 0x80);
      testWr (dev, 0xFE); testWr (dev, 0xFF);
      max = 1;
      break;
    case 0x8A:
      printf ("Assuming Conexant\n");
      printf ("Testing writes:\n");
      for (j = 0x00; j < 0x2e; j++) testWr (dev, j);
      testWr (dev, 0xFE); testWr (dev, 0xFF);
      max = 1;
      break;
    default:
      exit (1);
  }
#else
  max = 1;
#endif

  for (i = 0; i < max; i++) {
    printf ("\nTesting reads (%i)", i);
    if (chain->dev->SlaveAddr == 0x8A) {
      if (!i) {
//      printf (" (read enabled) ");
//      TVWriteBus (dev, 0x6c, 0x46); /* enable reg rd */
      } else {
        printf (" (read disabled) ");
        TVWriteBus (dev, 0x6c, 0x06); /* disable en_reg_rd */
      }
    }
    tvBusOk = TRUE;
    TVStatusBus (dev, &res);
    if (!tvBusOk) printf (" ?:--\n"); else printf (" ?:%02X\n", res);
    for (j = 0; j < 0x100; j++) {
      if ((j & 0xf) == 0x0) printf ("%02X:", j);
      tvBusOk = TRUE;
      TVReadBus (dev, j, &res);
      if (!tvBusOk) printf (" --"); else printf (" %02X", res);
      if ((j & 0xf) == 0xf) printf ("\n");
    }
    tvBusOk = TRUE;
    TVStatusBus (dev, &res);
    if (!tvBusOk) printf (" ?:--\n"); else printf (" ?:%02X\n", res);
  }

  back_card->getMode(&r);
  printf("clock %d\n", r.crtc.nv.clock);
  printf("HDisplay %d\n", r.crtc.nv.HDisplay);
  printf("HBlankStart %d\n", r.crtc.nv.HBlankStart);
  printf("HSyncStart %d\n", r.crtc.nv.HSyncStart);
  printf("HSyncEnd %d\n", r.crtc.nv.HSyncEnd);
  printf("HBlankEnd %d\n", r.crtc.nv.HBlankEnd);
  printf("HTotal %d\n", r.crtc.nv.HTotal);
  printf("VDisplay %d\n", r.crtc.nv.VDisplay);
  printf("VBlankStart %d\n", r.crtc.nv.VBlankStart);
  printf("VSyncStart %d\n", r.crtc.nv.VSyncStart);
  printf("VSyncEnd %d\n", r.crtc.nv.VSyncEnd);
  printf("VBlankEnd %d\n", r.crtc.nv.VBlankEnd);
  printf("VTotal %d\n", r.crtc.nv.VTotal);
  printf("latency %d\n", r.crtc.nv.latency);
  back_card->probeCard();
//  back_card->setMode (&r); 
  back_access->closeCard ();
}


void modifySet (TVSettings *s, OptIntVal *m)
{
  int i;

  while (m) {
    i = (char *) m->addr - (char *) s;
    if (i < 0 || i > sizeof(TVSettings)) {
      printf ("Fatal: Wrong settings to modify\n");
      exit (1);
    }
    *(m->addr) = m->val;
    m = m->link;
  }
}

Bool processOptSet (char *opt, OptIntVal **head)
{
  char *s;
  char *name;
  int val;
  OptIntDecl *p;
  OptIntVal *q;
  OptIntVal *tail = NULL;
  OptFlagDecl *r;

  if (!opt) return FALSE;
  name = opt;
  s = strchr(opt, ':');
  if (!s) {
    fprintf (stderr, "Illegal setting: %s\n", opt);
    return FALSE;
  }
  *s++ = '\0';
  DPRINTF ("process settings %s = %i\n", name, val);
  for (p = opt_set_list; p->name; p++) {
    if (strcasecmp (name, p->name) == 0) break;
  }
  if (p->name) {
    if (sscanf (s, "%i", &val) != 1) {
      fprintf (stderr, "Illegal setting: %s\n", opt);
      return FALSE;
    }
    if (val < p->min || val > p->max) {
      fprintf (stderr, "Setting %s out of range (%i to %i)\n", 
	       p->name, p->min, p->max);
      return FALSE;
    }
    q = (OptIntVal *) xalloc (sizeof (OptIntVal));
    q->addr = p->addr;
    q->link = NULL;
    q->val = val;
    if (tail) tail->link = q;
    tail = q;
    if (!(*head)) *head = q;
    return TRUE;
  }
  for (r = opt_flag_list; r->name; r++) {
    if (strcasecmp (name, r->name) == 0) break;
  }
  if (r->name) {
    if (strcasecmp (s, "false") == 0) val = 0; 
    else if (strcasecmp (s, "off") == 0) val = 0; 
    else if (strcasecmp (s, "0") == 0) val = 0; 
    else if (strcasecmp (s, "true") == 0) val = 1; 
    else if (strcasecmp (s, "on") == 0) val = 1; 
    else if (strcasecmp (s, "1") == 0) val = 1; 
    else {
      fprintf (stderr, "Illegal flag value: %s\n", s);
      return FALSE;
    }
    *(r->addr_mask) |= r->mask;
    if (val) {
      *(r->addr_val) |= r->mask;
    } else {
      *(r->addr_val) &= ~r->mask;
    }
    return TRUE;
  }
  fprintf (stderr, "Unknown setting %s\n", name);
  return FALSE;
}

int main (int argc, char *argv[])
{
  CardPtr main_card;
  int c = '?';
  ActionMain act_main = ACTION_NONE;
  ActionMain act_next;
  ActionAux act_aux = 0;
  Display *main_dpy;
  int main_screen;
  TVMode main_mode;
  TVRegs main_regs;
  int main_head;

  opterr = 0;
  while ((c = getopt_long (argc, argv, short_options, 
                long_options, NULL)) != EOF) 
  {
    act_next = ACTION_NONE;
    switch(c) 
    {
      case 'h': /* Print usage */
      case '?':
        usage(); 
	break;
      case 'D': 
	opt_dpy = optarg; 
	break;
      case '.': /* Gtk/Gdk option, ignore */
        break;
      case 'n':
	opt_no_root = TRUE; 
	break;
      case 'r':
	if (!optarg || sscanf (optarg, "%i,%i", &opt_res_x, &opt_res_y) != 2) 
	  usage ();
	break;
      case 'o':
	if (!optarg || sscanf (optarg, "%f,%f", &opt_hoc, &opt_voc) != 2)
	  usage ();
	break;
      case 'd':
	act_next = ACTION_DEBUG;
	break;
      case 'm':
	act_next = ACTION_TVOFF;
	break;
      case 't':
	act_next = ACTION_TVON;
	break;
      case 'b':
	act_next = ACTION_TVBARS;
	break;
      case 'p':
	act_next = ACTION_PRINT;
	break;
      case 'P':
	act_next = ACTION_PROBE;
	break;
      case 'g':
	act_next = ACTION_GUI;
	break;
      case 's':
	if (optarg) opt_size = optarg;
	break;
      case 'S':
	if (optarg && strcasecmp (optarg, "ntsc") == 0) 
	  opt_system = TV_SYSTEM_NTSC; 
	else if (optarg && strcasecmp (optarg, "ntsc-j") == 0) 
	  opt_system = TV_SYSTEM_NTSC_J; 
	else if (optarg && strcasecmp (optarg, "pal") == 0) 
	  opt_system = TV_SYSTEM_PAL; 
	else if (optarg && strcasecmp (optarg, "pal-60") == 0) 
	  opt_system = TV_SYSTEM_PAL_60; 
	else if (optarg && strcasecmp (optarg, "pal-n") == 0) 
	  opt_system = TV_SYSTEM_PAL_N; 
	else if (optarg && strcasecmp (optarg, "pal-nc") == 0) 
	  opt_system = TV_SYSTEM_PAL_NC; 
	else if (optarg && strcasecmp (optarg, "pal-m") == 0) 
	  opt_system = TV_SYSTEM_PAL_M; 
	else if (optarg && strcasecmp (optarg, "pal-m60") == 0) 
	  opt_system = TV_SYSTEM_PAL_M60; 
	else
	  usage ();
	break;
      case 'C':
	if (optarg && strcasecmp (optarg, "auto") == 0) 
	  opt_connect = CONNECT_AUTO; 
	else if (optarg && strcasecmp (optarg, "fbas") == 0) 
	  opt_connect = CONNECT_FBAS; 
	else if (optarg && strcasecmp (optarg, "svhs") == 0) 
	  opt_connect = CONNECT_SVHS; 
	else if (optarg && strcasecmp (optarg, "both") == 0) 
	  opt_connect = CONNECT_BOTH; 
	else if (optarg && strcasecmp (optarg, "convert") == 0) 
	  opt_connect = CONNECT_CONVERT; 
	else
	  usage ();
	break;
      case 'T':
	if (optarg && strcasecmp (optarg, "brooktree") == 0) 
	  opt_tv_chip = TV_BROOKTREE; 
	else if (optarg && strcasecmp (optarg, "conexant") == 0) 
	  opt_tv_chip = TV_CONEXANT; 
	else if (optarg && strcasecmp (optarg, "chrontel") == 0) 
	  opt_tv_chip = TV_CHRONTEL; 
	else if (optarg && strcasecmp (optarg, "philips") == 0) 
	  opt_tv_chip = TV_PHILIPS; 
	else
	  usage ();
	break;
      case 'w':
	opt_window_name = optarg;
	break;
      case 'W':
	if (sscanf (optarg, "0x%lx", &opt_window) != 1) opt_window = None;
	if (opt_window == None) {
	  if (sscanf (optarg, "%ld", &opt_window) != 1) opt_window = None;
	}
	if (opt_window == None)
	  fprintf (stderr, "Invalid window id format: %s\n.", optarg);
	break;
      case 'c':
        act_aux |= ACTION_CENTER;
        break;
      case 'X':
        act_aux |= ACTION_XMODE;
        break;
      case 'N':
        opt_nvdev = TRUE;
        break;
      case '1':
        opt_head = 1;
        break;
      case '2':
        opt_head = 2;
        break;
      case 300:
        if (!processOptSet (optarg, &opt_set_act))
	  usage ();
	break;
    }
    if (act_next != ACTION_NONE) 
    {
      if (act_main != ACTION_NONE) {
	fprintf (stderr, "nvtv: More than one action. Choosing last one.\n");
      }
      act_main = act_next;
    }
  }

  /* Do the probing right away, without processing any other options. */

  if (act_main == ACTION_PROBE) {
#ifdef DEBUG_PROBE    
    main_card = back_root_init ();
    if (!main_card) {
      fprintf (stderr, "Either you are not root, or no NVidia card found.\n");
      exit (1);
    }
    bdir_probeSystem (main_card);
    return 0;
#else
    fprintf (stderr, "No probing enabled in this built.\n");
    exit (1);
#endif
  }

  /* Start to process the options. First, if -s/-o is specified, default
     -r and -S, and assume -g if no main action is present. */

  if (opt_size || opt_hoc >= 0.0 || opt_voc >= 0.0) {
    if (opt_system == TV_SYSTEM_NONE) {
      fprintf (stderr, "Defaulting to PAL TV system.\n");
      opt_system = TV_SYSTEM_PAL;
    }
    if (opt_res_x <= 0) {
      opt_res_x = 800;
      fprintf (stderr, "Defaulting to x resolution %i\n", opt_res_x);
    }
    if (opt_res_y <= 0) {
      opt_res_y = 600;
      fprintf (stderr, "Defaulting to y resolution %i\n", opt_res_y);
    }
    if (act_main == ACTION_NONE) {
      act_main = ACTION_GUI;
    }
  }

  if (act_main == ACTION_NONE && !act_aux) {
    act_main = ACTION_GUI;
  }

  /* Next, choose the backend. If a main action is present, always try to 
     get a backend. If -n is specified, force the null backend. For
     printing, we can get away with having the null backend, as well. */

  if (act_main != ACTION_NONE) {
    if (opt_no_root) {
      DPRINTF ("init null\n");
      main_card = back_null_init (); 
    } else {
      if (back_root_avail ()) {
	DPRINTF ("init root\n");
	main_card = back_root_init ();
      } else if (back_client_avail ()) {
	DPRINTF ("init client\n");
	main_card = back_client_init ();
      } else if (opt_nvdev && back_nvdev_avail (TRUE)) {
	DPRINTF ("init nvdev\n");
	main_card = back_nvdev_init ();
      } else if (act_main == ACTION_PRINT) {
	DPRINTF ("init null\n");
	main_card = back_null_init ();
      } else {
	fprintf (stderr, "Cannot access cards. Either you are not root, "
                 "or the NVidia devices are\nnot accessible.\n");
	exit (1);
      }
    }
    if (!main_card) {
      fprintf (stderr, "No NVidia card found.\n");
      exit (1);
    }
  }

  /* Next, open the first card, unless the main action is none or the gui.
     Find the right chip, if specified. Note: If printing, we can find all 
     tv chip types in the null backend. */

  if (act_main != ACTION_NONE && act_main != ACTION_GUI && 
      act_main != ACTION_DEBUG) 
  {
    back_access->openCard (main_card);
    if (opt_head != -1 && 
	(act_main == ACTION_TVON || act_main == ACTION_TVOFF)) 
    {
      back_card->getHeads (&main_head, NULL, NULL, NULL);
      back_card->setHeads (-1, opt_head, opt_head);
      if (act_main == ACTION_TVON && opt_head != main_head &&
	  ! back_card->getTwinView (NULL, NULL)) 
      {
	back_card->initSharedView (NULL, NULL);
      }
    }
    if (opt_tv_chip != TV_NO_CHIP) {
      ChipPtr chip;

      for (chip = main_card->chips; chip; chip = chip->next) {
	if (opt_tv_chip == chip->chip) break;
      }
      if (chip) {
	back_card->setChip (chip, TRUE);
	/* FIXME: Should set head here */
      } else {
	fprintf (stderr, "Cannot find specified chip.\n");
	exit (1);
      }
    }
  }

  /* Next, find the mode for printing -p and tv-on -t. This is required,
     fail if no mode is found. */

  if (act_main == ACTION_TVON || act_main == ACTION_PRINT) {
    if (opt_size) {
      if (! back_card->findBySize (opt_system, opt_res_x, opt_res_y, opt_size,
				 &main_mode)) 
      {
	fprintf (stderr, "Cannot find '%s' mode %i x %i \n", 
		 opt_size, opt_res_x, opt_res_y);
	fprintf (stderr, "Please specify as e.g. -r 800,600 -s Large\n");
	exit (1);
      }
    } else if (opt_hoc >= 0.0 || opt_voc >= 0.0) {
      if (opt_hoc < 0.0) opt_hoc = 0.10;
      if (opt_voc < 0.0) opt_voc = 0.10;
      /* FIXME: Defaulting message */
      /* FIXME: findByOverscan */
      fprintf (stderr, "Specifying modes by overscan is not implemented yet\n");
      exit (1);
    } else {
      fprintf (stderr, "No mode specified.\n");
      fprintf (stderr, "Please specify as e.g. -r 800,600 -s Large\n");
      exit (1);
    }
  }
      
  /* For tv-off, the auxiliary actions or the gui, we need a display */

  if (act_main == ACTION_GUI || act_main == ACTION_TVOFF || act_aux) {
#if HAVE_GTK
    /* For some reason, just XOpenDisplay is not enough if GDK is 
       linked in / to be used later. */
    gdk_init (&argc, &argv);
    main_dpy = gdk_display;
    main_screen = gdk_screen;
#else
    main_dpy = XOpenDisplay (opt_dpy);
    main_screen = DefaultScreen (main_dpy);
#endif
  }

  /* For tv-off, we need a monitor mode. Get it with the XVidMode extension,
     and default resolution, if necessary. */

  if (act_main == ACTION_TVOFF) {
    if (opt_res_x > 0 && opt_res_y > 0) {
      find_vidmode (main_dpy, main_screen, opt_res_x, opt_res_y, 
		    &main_regs.crtc, data_card_func(main_card->type)->make);
    } else 
    if (!get_vidmode (main_dpy, main_screen, &opt_res_x, &opt_res_y, 
		      &main_regs.crtc, data_card_func(main_card->type)->make)) 
    {
      fprintf (stderr, "No monitor mode to switch tv off. "
	               "You need the XVidMode extension.\n");
      exit (1);
    }
    /* FIXME: Without XVidMode extension, try to get the mode from 
       somewhere else, i.e. internal VESA database? */
    /* NOTE: There must be a valid mode in main_crt here */
  }

  /* If there is still no resolution (i.e., opt_res both -1), then we cannot
     switch the X mode, even if requested, so disable that */

  if ((act_aux & ACTION_XMODE) && ((opt_res_x < 0) || (opt_res_y < 0))) {
    act_aux &= ~ACTION_XMODE;
    fprintf (stderr, "No resolution found. Switching X mode disabled.\n");
  }

  /* Finally, modify the settings, unless the main action is none or gui. */

  if (act_main != ACTION_NONE && act_main != ACTION_GUI &&
      act_main != ACTION_DEBUG && back_card) 
  {
    back_card->getSettings (&opt_set);
    if (opt_connect > CONNECT_NONE) {
      opt_set.connector = opt_connect;
    } else {
      opt_set.connector = CONNECT_BOTH;
    }
    modifySet (&opt_set, opt_set_act);
    opt_set.flags = opt_mode_flags;
    /* FIXME: Should probe if CONNECT_AUTO */
  }

  /* Now we start executing the actions. Printing, color bars and 
     debugging ignore the auxiliary actions, and are handled first. */

  switch (act_main) 
  {
    case ACTION_TVBARS:
      back_card->setTestImage (NULL, &opt_set);
      return 0;
    case ACTION_PRINT:
      printRegs (&main_mode, main_card->type);
      return 0;
    case ACTION_DEBUG:
      doDebug ();
      return 0;
    default:
      break;
  }
  
  /* If we reached this point, we have to process the auxiliary options.
     If the action is tv-off, we should execute the action first, so
     switching the X mode comes afterwards. There must be valid monitor
     mode in main_crt at this point. */

  if (act_main == ACTION_TVOFF) {
    back_card->setMode (&main_regs);
  }

  /* Now, switch the X mode */

  if (act_aux & ACTION_XMODE) {
    switch_vidmode (main_dpy, main_screen, opt_res_x, opt_res_y);
    XSync (main_dpy, False);
  }

  /* Next, look for the window. */

  if (opt_window_name) {
    opt_window = Window_With_Name (main_dpy, 
      RootWindow(main_dpy, main_screen), opt_window_name);
    if (opt_window == None) {
      fprintf (stderr, "No window with name '%s' exists.\n", opt_window_name);
    }
  }

  /* Next, center the window, with respect to the resolution, if given. */

  if (act_aux & ACTION_CENTER) {
    if (opt_window != None) {
      center_window (main_dpy, main_screen, opt_window, opt_res_x, opt_res_y);
    } else {
      fprintf (stderr, "No window specified.\n");
      exit (1);
    }
  }

  /* Next, switch tv mode on. The main_mode must be a valid tv mode at
     this point. */

  if (act_main == ACTION_TVON) {
    main_mode.regs.devFlags = DEV_MONITOR | DEV_TELEVISION;

    if (opt_mode_mask & TV_DESC_DUALVIEW) {
      if (! (opt_mode_flags & TV_DESC_DUALVIEW))
	main_mode.regs.devFlags = DEV_TELEVISION;
    }
    back_card->setModeSettings (&main_mode.regs, &opt_set);
  }

  /* Finally, start the gui. */

  if (act_main == ACTION_GUI) {
#ifdef HAVE_GTK
    gui_main (argc, argv, main_card); 
#else
    fprintf (stderr, "Compiled without gui\n");
    return 1;
#endif
  }

  return 0;
}

/* -------------------------------------------------------------------

Table of options/action:

                                       mode  backend   card    head
none     -r -X          -c             no    no        ignore  no
gui  -g  -r -X -s/-o -S -c -C/-f       opt   yes       ignore  no
on   -t  -r -X -s/-o -S -c -C/-f -1/2  yes   yes       open    yes
off  -m  -r -X          -c       -1/2  no    yes       open    yes
bars -b                    -C/-f       no    yes       open    no
prt  -p  -r    -s/-o -S 	       yes   yes/none  open    no
     -P
     -d


Rules:

If only -s/-o is given, -r and -S are defaulted, regardless of action.
If -s/-o is specified and no action given, assume -g.
For -g, only auxiliary options are executed. No mode search, no settings.
Mode search is only done for -t and -p, and aborts if none is found/specified.
If card is opened, --chip will be respected.

----------------------------------------------------------------------


xine --display-ratio 0.979 # for 768x576 Large

Not implemented yet:

-x   --x-options         standard x options follow

GDK Options:

+ --display
+ --sync
+ --no-xshm
+ --name
+ --class
  --gxid_host
  --gxid_port
  --xim-preedit
  --xim-status

GTK Options

  --gtk-module
  --g-fatal-warnings

*/


