/* NVTV TV common routines -- Dirk Thierbach <dthierbach@gmx.de>
 *
 * This file is part of nvtv, a tool for tv-output on NVidia cards.
 * 
 * nvtv is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * nvtv is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: tv_common.c,v 1.1.1.1 2003/01/02 04:49:41 rshoff Exp $
 *
 * Contents:
 *
 * Header: Common tv-related routines.
 *
 */

#include <stdarg.h>
#include <string.h>

#include "debug.h"
#include "xfree.h" /* for NULL, xcalloc, ... */
#include "xf86i2c.h"
#include "tv_common.h"
#include "tv_bt.h"
#include "tv_cx.h"
#include "tv_ch.h"
#include "tv_ph.h"
#include "tv_null.h"

/* -------- */

TVState tvState = TV_UNKNOWN;

/* -------- */

I2CChainPtr TVFindDevice (I2CChainPtr root, TVChip chip)
{
  I2CChainPtr chain;

  for (chain = root; chain; chain = chain->next)
  {
    if (chain->chip == chip) return chain;
  }
  return NULL;
}

/* FIXME: NVFindTvDevice -> TVFindDevice (pNv->TvChain, ...) */

/* 
 * Create chain of tv chips from list of devices on busses
 */

I2CChainPtr TVCreateChain (I2CBusPtr busses[], int nbus)
{
  int bus;
  I2CDevPtr dev;
  I2CChainPtr chain, root;
  char version[50];
  char *s;

  root = NULL;
  for (bus = nbus; bus > 0; ) {
    bus--;
    for (dev = busses[bus]->FirstDev; dev; dev = dev->NextDev) {
      chain = xcalloc (1, sizeof (I2CChainRec));
      chain->dev = dev;
      chain->next = root;
      chain->chip = TV_NO_CHIP;
      switch (dev->SlaveAddr) {
	case 0x88: 
        case 0x8a: /* FIXME: For test only */
	  if ((s = TVDetectConexant (dev, &chain->chip))) break; 
	  if ((s = TVDetectPhilips (dev, &chain->chip))) break; 
	  /* no break */
	/* case 0x8a: FIXME */
	  if ((s = TVDetectBrooktree (dev, &chain->chip, tvState))) break; 
	  break;
	case 0xea:
	case 0xec:
	  if ((s = TVDetectChrontel (dev, &chain->chip))) break; 
	  break;
      }
      if (chain->chip == TV_NO_CHIP) {
	snprintf (version, 50, "Unknown chip (%1s:%02X)", I2C_ID(dev));
	s = version;
      }
      chain->name = xalloc (strlen(s)+1);
      strcpy (chain->name, s);
      root = chain;
    }
  }
  return root;
}

/*
 * Free the chain of tv chips
 */

void TVDestroyChain (I2CChainPtr root)
{
  I2CChainPtr chain;

  while (root) {
    xfree (root->name);
    chain = root->next;
    xfree (root);
    root = chain;
  }
}

void TVProbeDevice (I2CBusPtr bus, I2CSlaveAddr addr, char *format, ...)
{
  I2CDevPtr dev;
  char *s;
  va_list ap;

#ifndef FAKE_PROBE_ALL
#ifndef FAKE_PROBE_ADDR
  if (xf86I2CProbeAddress(bus, addr))
#else
  if (addr == FAKE_PROBE_ADDR)
#endif
#endif
  {
    dev = xf86CreateI2CDevRec();
    s = xalloc (8); 
    va_start (ap, format);
    vsnprintf (s, 7, format, ap);
    va_end (ap);
    dev->DevName = s;
    dev->SlaveAddr = addr;
    dev->pI2CBus = bus;
    if (!xf86I2CDevInit(dev)) {
      xfree (dev->DevName); 
      xf86DestroyI2CDevRec(dev, TRUE);
    }
  }
}

void TVProbeKnownDevices (I2CBusPtr busses[], int nbus)
{
  int bus;
  
  for (bus = 0; bus < nbus; bus++) {
    TVProbeDevice (busses[bus], 0x88, "%i:%02X", bus, 0x88);
    TVProbeDevice (busses[bus], 0x8A, "%i:%02X", bus, 0x8A);
    TVProbeDevice (busses[bus], 0xEA, "%i:%02X", bus, 0xEA);
    TVProbeDevice (busses[bus], 0xEC, "%i:%02X", bus, 0xEC);
  }
}

void TVProbeAllDevices (I2CBusPtr busses[], int nbus)
{
  I2CSlaveAddr addr;
  int bus;

  for (bus = 0; bus < nbus; bus++) {
    for (addr = 0x00; addr < 0x100; addr += 2) {
      TVProbeDevice (busses[bus], addr, "%1i:%02X", bus, addr);
    }
  }
}

void TVDestroyDevices (I2CBusPtr busses[], int nbus)
{
  I2CDevPtr dev;
  int bus;

  for (bus = 0; bus < nbus; bus++) {
    if (busses[bus]) {
      while ((dev = busses[bus]->FirstDev) != NULL) {
	xfree (dev->DevName);
	xf86DestroyI2CDevRec(dev, TRUE); 
      }
    }
  }
}

void TVDestroyBusses (I2CBusPtr busses[], int nbus)
{
  int i;

  for (i = 0; i < nbus; i++) 
  {
    if (busses[i]) xf86DestroyI2CBusRec (busses[i], TRUE, FALSE);
    busses[i] = NULL;
  }
}

void TVSetTvEncoder (TVEncoderObj *encoder, I2CChainPtr chain)
{
  switch (chain->chip & TV_ENCODER) {
    case TV_BROOKTREE:
      *encoder = tvBtTemplate;
      break;
    case TV_CONEXANT:
      *encoder = tvCxTemplate;
      break;
    case TV_CHRONTEL:
      *encoder = tvChTemplate;
      break;
    case TV_PHILIPS:
      *encoder = tvPhTemplate;
      break;
    default:
      *encoder = tvNullTemplate;
      break;
  }
  encoder->Create (encoder, chain->chip, chain->dev);
}

